package com.zh.lingnuo.utils;

import com.zh.lingnuo.annotation.ExcelParam;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.ss.formula.functions.T;
import org.apache.poi.ss.usermodel.*;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.web.multipart.MultipartFile;

import java.beans.IntrospectionException;
import java.beans.PropertyDescriptor;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * 导入的方法
 * @author chenliuping
 * @date 2021-06-24 11:06
 */
public class ExcelRead {
    //日志输出
    //private static Logger logger = LoggerFactory.getLogger(ExcelRead.class);
    //定义excel类型
    private static final String XLS = "xls";
    private static final String XLSX = "xlsx";

    /**
     * 根据文件后缀名类型获取对应的工作簿对象
     * @param inputStream 读取文件的输入流
     * @param fileType 文件后缀名类型（xls或xlsx）
     * @return 包含文件数据的工作簿对象
     */
    private static Workbook getWorkbook(InputStream inputStream, String fileType) throws IOException {
        //用自带的方法新建工作薄
        //Workbook workbook = WorkbookFactory.create(inputStream);
        //后缀判断有版本转换问题
        Workbook workbook = null;
        if (fileType.equalsIgnoreCase(XLS)) {
            workbook = new HSSFWorkbook(inputStream);
        } else if (fileType.equalsIgnoreCase(XLSX)) {
            workbook = new XSSFWorkbook(inputStream);
        }
        return workbook;
    }


    /**
     * <p>将单元格内容转换为字符串</p>
     * @param cell
     * @return
     */
    private static String convertCellValueToString(Cell cell) {
        if (cell == null) {
            return null;
        }
        String returnValue = null;
        switch (cell.getCellType()) {
            case NUMERIC: //数字
                Double doubleValue = cell.getNumericCellValue();
                // 格式化科学计数法，取一位整数，如取小数，值如0.0,取小数点后几位就写几个0
                DecimalFormat df = new DecimalFormat("0");
                returnValue = df.format(doubleValue);
                break;
            case STRING: //字符串
                returnValue = cell.getStringCellValue();
                break;
            case BOOLEAN: //布尔
                Boolean booleanValue = cell.getBooleanCellValue();
                returnValue = booleanValue.toString();
                break;
            case BLANK: // 空值
                break;
            case FORMULA: // 公式
                returnValue = cell.getCellFormula();
                break;
            case ERROR: // 故障
                break;
            default:
                break;
        }
        return returnValue;
    }

    /**
     * <p>根据给定的实体类中赋值的注解值读取Excel</p>
     * <p>fileName:Excel文件路径</p>
     * <p>StatrRow：读取的开始行数（默认填0）</p>
     * <p>EndRow：读取的结束行数（填-1为全部）</p>
     * <p>sheetNums：导入第几个工作表（填-1为全部）</p>
     * <p>Class<T>：传过来的实体类类型</p>
     * <p>返回一个List<?>:?为实体类</p>
     * @return
     */
    public static List<?> ReadExcelByAll(MultipartFile file, int StatrRow, int EndRow, int sheetNums, Class t) throws InvocationTargetException, IntrospectionException, InstantiationException, IllegalAccessException, NoSuchFieldException {
        //判断输入的开始值是否少于等于结束值
        if (StatrRow > EndRow && EndRow != -1) {
            /*logger.warn("输入的开始行值比结束行值大，请重新输入正确的行数");
            List<T> error = null;
            return error;*/
            throw new RuntimeException("输入的开始行值比结束行值大，请重新输入正确的行数");
        }
        if(sheetNums < -1){
            List<T> error = null;
            return error;
        }
        //声明返回的结果集
        List<T> result = new ArrayList<>();
        //声明一个工作薄
        Workbook workbook = null;
        //声明一个文件输入流
        InputStream inputStream =null;
        String fileName = file.getOriginalFilename();
        try {
            String fileType = fileName.substring(fileName.lastIndexOf(".") + 1);
            inputStream = file.getInputStream();
            workbook = getWorkbook(inputStream, fileType);
            //处理Excel内容
            result = HandleData(workbook, StatrRow, EndRow,sheetNums, t);
        } catch (Exception e) {
            //logger.warn("解析Excel失败，文件名：" + fileName + " 错误信息：" + e.getMessage());
            throw new RuntimeException("解析Excel失败，文件名：" + fileName + " 错误信息：" + e.getMessage());
        } finally {
            try {
                if (null != workbook) {
                    workbook.close();
                }
                if (null != inputStream) {
                    inputStream.close();
                }
            } catch (Exception e) {
                /*//logger.warn("关闭数据流出错！错误信息：" + e.getMessage());
                return null;*/
                throw new RuntimeException("关闭数据流出错！错误信息：" + e.getMessage());
            }
        }
        return result;
    }

    /**
     * <p>处理Excel内容转为List<T>输出</p>
     * <p>workbook：已连接的工作薄</p>
     * <p>StatrRow：读取的开始行数（默认填0，0开始,传过来是EXcel的行数值默认从1开始，这里已处理减1）</p>
     * <p>EndRow：读取的结束行数（填-1为全部）</p>
     * <p>sheetNums：导入第几个工作表（填-1为全部）</p>
     * <p>Class<T>：所映射的实体类</p>
     */
    private static <T> List<T> HandleData(Workbook workbook, int StatrRow, int EndRow,int sheetNums, Class<?> t) throws IntrospectionException, NoSuchFieldException, IllegalAccessException, InstantiationException, InvocationTargetException, ClassNotFoundException {
        //声明返回的结果集
        List<T> result = new ArrayList<>();
        //判断导入的工作表
        int startSheet = 0;
        int totalSheet = workbook.getNumberOfSheets();
        if (sheetNums == -1){
            startSheet = 0;
        }else if(sheetNums == 0){
            return null;
        }else{
            startSheet = sheetNums - 1;
            totalSheet = sheetNums;
        }
        //解析sheet（sheet是Excel脚页）
        //循环工作表,特殊处理，不会影响速度
        for (int sheetNum = startSheet; sheetNum < totalSheet; sheetNum++) {
            Sheet sheet = workbook.getSheetAt(sheetNum);
            // 校验sheet是否合法
            if (sheet == null) {
                continue;
            }
            //获取头部数据
            //声明头部数据数列对象
            ArrayList<String> top = new ArrayList<>();
            //获取Excel第一行数据
            int firstRowNum = sheet.getFirstRowNum();
            Row firstRow = sheet.getRow(firstRowNum);
            if (null == firstRow) {
                //logger.warn("解析Excel失败，在第一行没有读取到任何数据！");
                continue;
            }
            for (int i = 0; i < firstRow.getLastCellNum(); i++) {
                top.add(convertCellValueToString(firstRow.getCell(i)));
            }
            //获取实体类的成原变量
            Map<String, Object> entityFields = getFieldAndValue(t);
            //判断所需要的数据列
            Map<String, Object> exceltoEntity = new HashMap<>();
            for (int i = 0; i < top.size(); i++) {
                if (entityFields.get(top.get(i)) != null && !"".equals(entityFields.get(top.get(i)))) {
                    exceltoEntity.put(String.valueOf(i), entityFields.get(top.get(i)));
                }
            }
            /*处理Excel数据内容*/
            int endRowNum;
            //获取结束行数
            if (EndRow == -1) {
                endRowNum = sheet.getPhysicalNumberOfRows();
            } else {
                endRowNum = EndRow <= sheet.getPhysicalNumberOfRows() ? EndRow : sheet.getPhysicalNumberOfRows();
            }
            List<Map<String, Object>> mapList = new ArrayList<>();
            //遍历行数
            for (int i = StatrRow - 1; i < endRowNum; i++) {
                Row row = sheet.getRow(i);
                if (null == row) {
                    continue;
                }
                //获取需要的列数据
                T texcel = (T) t.newInstance();
                for (Map.Entry<String, Object> map : exceltoEntity.entrySet()) {
                    //获取Exceld对应列的数据
                    String celldata = convertCellValueToString(row.getCell(Integer.parseInt(map.getKey())));
                    //使用发射
                    //获取实体类T中指定成员变量的对象
                    PropertyDescriptor pd = new PropertyDescriptor((String) map.getValue(), texcel.getClass());
                    //获取成员变量的set方法
                    Method method = pd.getWriteMethod();
                    //判断成员变量的类型
                    Field field = texcel.getClass().getDeclaredField((String) map.getValue());
                    String object = field.getGenericType().getTypeName();
                    System.out.println(object.toString());
                    if (object.endsWith("String")) {
                        //执行set方法
                        method.invoke(texcel, celldata);
                    }
                    if (object.endsWith("Double")) {
                        Double middata = Double.valueOf(celldata);
                        //执行set方法
                        method.invoke(texcel, middata);
                    }
                    if (object.endsWith("Float")) {
                        Float middata = Float.valueOf(celldata);
                        //执行set方法
                        method.invoke(texcel, middata);
                    }
                    if (object.endsWith("Integer")) {
                        Integer middata = null;
                        try {
                            middata = Integer.parseInt(celldata);
                        } catch (NumberFormatException e) {

                        }
                        //执行set方法
                        method.invoke(texcel, middata);
                    }
                }
                result.add(texcel);
            }
        }
        return result;
    }

    /**
     * 获取对应的实体类成员
     * */
    private static Map<String, Object> getFieldAndValue(Class T) {
        //声明返回结果集
        Map<String, Object> result = new HashMap<>();
        Field[] fields = T.getDeclaredFields();//获取属性名
        if (fields != null) {
            for (Field field : fields) {
                ExcelParam param = field.getAnnotation(ExcelParam.class);
                if (param.value() != null && !"".equals(param.value())) {
                    result.put(param.value(), field.getName());
                }
            }
        } else {
            //logger.warn("实体类：" + T + "不存在成员变量");
            //return null;
            throw new RuntimeException("实体类：" + T + "不存在成员变量");
        }
        return result;
    }
}
